
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t limit,char *pv),
		ReadSliceImage(char *path,int X,int Y,Cell **cell);

#ifndef	UMA
#define UMA	1024	/* unit of memory allocation */
#endif

static long	*Link;

static long	Trace(long old)
{
	long	new=Link[old];

	return (new<0)?old:(Link[old]=Trace(new));
}

typedef struct {
		long	l_p;
		int	x1,y1,z1,x2,y2,z2;
	} Cluster;

static int	Compare(Cluster *C1,Cluster *C2)
{
	long	p1=Link[C1->l_p],p2=Link[C2->l_p];

	return (p1==p2)?0:(p1<p2)?-1:1;
}

#define PV	(Cell)(~0)
#define LEN	1024
#define SOC	sizeof(Cell)
#define BPS	(SOC*8)
#define SOL	sizeof(long)

int main(int argc,char **argv)
{
	char	**paths,pv[(size_t)PV+1],path[LEN],
		*mae="memory allocation error.",
		*tlc="too large cluster.",
		*mrae="memory re-allocation error.";
	int	X,Y,Z,z,y,x,len,bps;
	Cell	**cell;
	long	**last,L,l,m,N;
	size_t	S;
	Cluster	*C;

#ifdef	IOM
	Cell	***B,**b,mask;
#endif
	if (argc!=4 && argc!=5)
	    Error("usage : si_mcl_1 directory nameFile rangeList {directory}");

	paths=CheckSliceImages(argv[1],argv[2],&X,&Y,&Z,NULL);

	RangeList(argv[3],PV,pv);

#ifdef	IOM
{
	int	i=(Z+BPS-1)/BPS;

	if ((B=(Cell ***)malloc(sizeof(Cell **)*i))==NULL ||
	    (*B=(Cell **)malloc(sizeof(Cell *)*i*Y))==NULL ||
	    (**B=(Cell *)malloc(SOC*i*Y*X))==NULL) Error(mae);

	for (z=0; z<i; z++) {
	    if (z>0) B[z]=B[z-1]+Y;

	    b=B[z];
	    for (y=0; y<Y; y++) {
		if (z>0 || y>0) b[y]=b[y-1]+X;

		for (x=0; x<X; x++) b[y][x]=0;
	    }
	}
}
#endif
	if ((cell=(Cell **)malloc(sizeof(Cell *)*Y))==NULL ||
            (*cell=(Cell *)malloc(SOC*Y*X))==NULL) Error(mae);

	for (y=1; y<Y; y++) cell[y]=cell[y-1]+X;

#ifdef	IOM
#define READ(z)		b=B[z/BPS]; mask=(Cell)(1<<z%BPS)
#define TEST(x,y)	b[y][x]&mask

	for (z=0; z<Z; z++) {
	    ReadSliceImage(paths[z],X,Y,cell);

	    READ(z);
	    for (y=0; y<Y; y++)
	    for (x=0; x<X; x++) if (!pv[cell[y][x]]) b[y][x]|=mask;
	}
	if (argc!=5) {
	    free(*cell); free(cell);
	}
#else
#define READ(z)		ReadSliceImage(paths[z],X,Y,cell)
#define TEST(x,y)	!pv[cell[y][x]]
#endif
	if ((last=(long **)malloc(sizeof(long *)*(Y+1)))==NULL ||
	    (*last=(long *)malloc(SOL*(Y*(X+1)+X)))==NULL ||
	    (Link=(long *)malloc(SOL*(S=UMA)))==NULL) Error(mae);

	++last; for (y=0; y<Y; y++) last[y]=last[y-1]+X+1;

	for (x=0; x<X; x++) last[-1][x]=0;
	for (y=0; y<Y; y++) last[y][-1]=0;

	Link[0]=0;

	L=1;
	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) last[y][x]=0;

	for (z=0; z<Z; z++) {
	    READ(z);

#define CHECK(size)     if ((size)>=0) Error(tlc)

	    for (y=0; y<Y; y++)
	    for (x=0; x<X; x++) {
		if (TEST(x,y)) {
		    CHECK(--Link[l=0]);
		}
		else if ((l=last[y][x])!=0) {
		    CHECK(--Link[l=Trace(l)]);
		    if ((m=last[y-1][x])!=0) {
			if ((m=Trace(m))!=l) {
			    CHECK(Link[l]+=Link[m]); Link[m]=l;
		 	}
			if ((m=last[y][x-1])!=0 && (m=Trace(m))!=l) {
			    CHECK(Link[l]+=Link[m]); Link[m]=l;
			}
		    }
		    else if ((m=last[y][x-1])!=0 && m!=l) {
			CHECK(Link[l]+=Link[m]); Link[m]=l;
		    }
		}
		else if ((l=last[y-1][x])!=0) {
		    CHECK(--Link[l=Trace(l)]);
		    if ((m=last[y][x-1])!=0 && m!=l) {
			CHECK(Link[l]+=Link[m]); Link[m]=l;
		    }
		}
		else if ((l=last[y][x-1])!=0) {
		    CHECK(--Link[l]);
		}
		else {
		    if (L==S && (Link=(long *)
				 realloc(Link,SOL*(S+=UMA)))==NULL) Error(mrae);

		    Link[l=L++]=(-1);
		}
		last[y][x]=l;
	    }
	}
	if (L<S && (Link=(long *)realloc(Link,SOL*L))==NULL) Error(mrae);

	for (N=l=1; l<L; l++) if ((m=Link[l])<0) ++N; else Link[l]=Trace(m);

	if ((C=(Cluster *)malloc(sizeof(Cluster)*N))==NULL) Error(mae);

	for (N=l=1; l<L; l++) if (Link[l]<0) C[N++].l_p=l;

	if (N>1) qsort(C+1,N-1,sizeof(Cluster),(int (*)())Compare);

	C[0].l_p=0;
	for (m=0; m<N; m++) {
	    C[m].l_p=Link[l=C[m].l_p]; Link[l]=(-m);

	    C[m].x1=X; C[m].y1=Y; C[m].z1=Z; C[m].x2=C[m].y2=C[m].z2=(-1); 
	}
	for (l=1; l<L; l++) if ((m=Link[l])>0) Link[l]=(-Link[m]);
	for (l=1; l<L; l++) if ((m=Link[l])<0) Link[l]=(-m);

	if (argc==5) {
	    for (len=0; argv[1][len]!='\0'; len++) ; ++len;

	    for (bps=1; bps<BPS && (size_t)1<<bps<N; bps<<=1) ;
	}
	L=1;
	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) last[y][x]=0;

	for (z=0; z<Z; z++) {
	    READ(z);

	    for (y=0; y<Y; y++)
	    for (x=0; x<X; x++) {
		if (TEST(x,y))
		    l=0;
		else if ((l=last[y][x])==0 && (l=last[y-1][x])==0 &&
					      (l=last[y][x-1])==0)
		    l=L++;

		m=Link[l]; if (argc==5) cell[y][x]=(m<PV)?m:PV;

		if (C[m].x1>x) C[m].x1=x; if (C[m].x2<x) C[m].x2=x;
		if (C[m].y1>y) C[m].y1=y; if (C[m].y2<y) C[m].y2=y;
		if (C[m].z1>z) C[m].z1=z; if (C[m].z2<z) C[m].z2=z;

		last[y][x]=l;
	    }
	    if (argc==5) {
		(void)sprintf(path,"%s/%s",argv[4],paths[z]+len);
		StoreImageFile(path,X,Y,bps,cell,NULL);
	    }
	}
	(void)printf("%ld\n",N);
	for (m=0; m<N; m++)
	    (void)printf("%lu\t%d\t%d\t%d\t%d\t%d\t%d\n",
			 (unsigned long)(-(long long)C[m].l_p),
			 C[m].x1,C[m].y1,C[m].z1,C[m].x2,C[m].y2,C[m].z2);

	return 0;
}
