
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

#ifndef	LABEL_T
#define LABEL_T	unsigned
#endif
#ifndef	PELS_T
#define PELS_T	unsigned long
#endif
#ifndef	XYZ_T
#define XYZ_T	unsigned short
#endif

extern void	Error(),RangeList(),ReadSliceImage();

static LABEL_T	Trace(link,old)
LABEL_T		*link,old;
{
	LABEL_T	new=link[old];

	return (new==0)?old:(link[old]=Trace(link,new));
}

typedef struct {
		PELS_T	pels;
		union {
			LABEL_T	l;
			struct {
				LABEL_T	old,new;
			} LABEL;
			struct {
				XYZ_T	x1,y1,z1,x2,y2,z2;
			} XYZ;
		} UNION;
	} Cluster;

static int	Compare(C1,C2)
Cluster		*C1,*C2;
{
	return (C1->pels==C2->pels)?0:(C1->pels>C2->pels)?-1:1;
}

#define PV	(Cell)(~0)
#define LEN	1024
#define BPS	(sizeof(Cell)<<3)

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,pv01[(size_t)PV+1],path[LEN];
	int	Nx,Ny,Nz,x,y,z,bps,noa=1;
	Cell	**cell;
	LABEL_T	**last,Nl,l,*link,lyx,Nc,c;
	Cluster	*cluster,*C;
	size_t	len;

#ifdef	IOM
	int	Z;
	Cell	***image,**block,mask;
#endif
	if (argc!=4 && argc!=5)
	    Error("usage : si_mhl directory nameFile rangeList {directory}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],(size_t)PV,pv01);

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

#ifdef	IOM
	Z=(Nz+BPS-1)/BPS;

	if ((image=ALLOC(Cell **,Z))==NULL ||
	    (*image=ALLOC(Cell *,(size_t)Z*(size_t)Ny))==NULL ||
	    (**image=ALLOC(Cell,(size_t)Z*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for image.");

	cell=image[0]; for (y=1; y<Ny; y++) cell[y]=cell[y-1]+Nx;

	for (z=1; z<Z; z++) {
	    cell=image[z]=image[z-1]+Ny;
	    for (y=0; y<Ny; y++) {
		cell[y]=image[z-1][y]+(size_t)Ny*(size_t)Nx;

		for (x=0; x<Nx; x++) cell[y][x]=0;
	    }
	}
#endif
	if ((cell=ALLOC(Cell *,Ny))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for cell.");

	for (y=1; y<Ny; y++) cell[y]=cell[y-1]+Nx;

#ifdef	IOM
	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++)
		if (pv01[cell[y][x]]) block[y][x]|=mask;
	}
	if (argc!=5) {
	    free(*cell); free(cell);
	}

#define READ(z)		block=image[z/BPS]; mask=(Cell)(1<<z%BPS)
#define TEST(x,y)	block[y][x]&mask
#else
#define READ(z)		ReadSliceImage(paths[z],Nx,Ny,cell)
#define TEST(x,y)	pv01[cell[y][x]]
#endif
	if ((last=ALLOC(LABEL_T *,1+Ny))==NULL ||
	    (*last=ALLOC(LABEL_T,(size_t)Nx+(size_t)Ny*(size_t)(1+Nx)))==NULL)
	    Error("no allocatable memory for last.");

	++last;
	for (x=0; x<Nx; x++) last[-1][x]=1;
	for (y=0; y<Ny; y++) {
	    last[y]=last[y-1]+1+Nx; for (x=(-1); x<Nx; x++) last[y][x]=1;
	}
	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if (TEST(x,y))
		    l=noa=0;
		else {
		    if (!last[y][x] && !last[y-1][x] && !last[y][x-1] &&
			++Nl>(LABEL_T)(~0)/sizeof(LABEL_T))
			Error("too many links.");

		    l=1;
		}
		last[y][x]=l; 
	    }
	}
	if (noa) Error("no object area.");

	if ((link=ALLOC(LABEL_T,Nl))==NULL)
	    Error("no allocatable memory for link.");

	link[0]=link[1]=0;

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) last[y][x]=1;

	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    if (TEST(x,y))
			l=0;
		    else if ((l=last[y][x])!=0) {
			l=Trace(link,l);
			if ((lyx=last[y-1][x])!=0 && (lyx=Trace(link,lyx))!=l)
			    if (lyx==1) l=link[l]=1; else link[lyx]=l;

			if ((lyx=last[y][x-1])!=0 && (lyx=Trace(link,lyx))!=l)
			    if (lyx==1) l=link[l]=1; else link[lyx]=l;
		    }
		    else if ((l=last[y-1][x])!=0) {
			l=Trace(link,l);
			if ((lyx=last[y][x-1])!=0 && lyx!=l)
			    if (lyx==1) l=link[l]=1; else link[lyx]=l;
		    }
		    else if ((l=last[y][x-1])==0)
			link[l=Nl++]=0;

		    last[y][x]=l;
		}
		if (z>0 && y>0 && Nx>1 && (l=last[y][Nx-1])>1) link[l]=1;
	    }
	    if (z>0 && Ny>1)
		for (x=Nx-2; x>=1; x--)
		    if ((l=last[Ny-1][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	}
	if (Nz>1)
	    for (y=Ny-2; y>=1; y--) for (x=Nx-2; x>=1; x--)
		if ((l=last[y][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;

	for (Nc=l=2; l<Nl; l++)
	    if (link[l]!=0)
		link[l]=Trace(link,link[l]);
	    else
		if (++Nc>(LABEL_T)(~0)/sizeof(Cluster))
		    Error("too many clusters.");

	(void)printf("%" LL_FORMAT "u\n",(unsigned long long)Nc);

	if ((cluster=ALLOC(Cluster,Nc))==NULL)
	    Error("no allocatable memory for cluster.");

#define L	UNION.l

	for (Nc=l=0; l<Nl; l++)
	    if (link[l]==0) {
		C=cluster+(link[l]=Nc++); C->pels=0; C->L=l;
	    }

	for (C=cluster+Nc-1, l=Nl-1; l>1; l--)
	    if (l==C->L) --C; else link[l]=link[link[l]];

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) last[y][x]=1;

	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if (TEST(x,y))
		    c=0;
		else if ((c=last[y  ][x  ])==0 &&
			 (c=last[y-1][x  ])==0 &&
			 (c=last[y  ][x-1])==0) c=link[Nl++];

		if ((last[y][x]=c)<=1) c=1-c;

		if (++cluster[c].pels==0) --cluster[c].pels;
	    }
	}

#define OLD	UNION.LABEL.old
#define NEW	UNION.LABEL.new

	if (Nc>2) {
	    for (C=cluster+(c=2); c<Nc; c++, C++) C->OLD=c;

	    qsort(cluster+2,(size_t)(Nc-2),sizeof(Cluster),Compare);

	    for (C=cluster+(c=2); c<Nc; c++, C++) cluster[C->OLD].NEW=c;

	    for (l=2; l<Nl; l++)
		if (link[l]!=1) link[l]=cluster[link[l]].NEW;
	}

#define X1	UNION.XYZ.x1
#define Y1	UNION.XYZ.y1
#define Z1	UNION.XYZ.z1
#define X2	UNION.XYZ.x2
#define Y2	UNION.XYZ.y2
#define Z2	UNION.XYZ.z2

	for (C=cluster, c=0; c<Nc; c++, C++) {
	    C->X1=(XYZ_T)(Nx-1);
	    C->Y1=(XYZ_T)(Ny-1);
	    C->Z1=(XYZ_T)(Nz-1); C->X2=C->Y2=C->Z2=0;
	}
	if (argc==5) {
	    len=strlen(argv[1])+1;

	    for (bps=1; bps<BPS && 1<<bps<Nc; bps<<=1) ;
	}
	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) last[y][x]=1;

	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if (TEST(x,y))
		    c=0;
		else if ((c=last[y  ][x  ])==0 &&
			 (c=last[y-1][x  ])==0 &&
			 (c=last[y  ][x-1])==0) c=link[Nl++];

		if ((last[y][x]=c)<=1) c=1-c;

		C=cluster+c;
		if (C->X1>x) C->X1=(XYZ_T)x; if (C->X2<x) C->X2=(XYZ_T)x;
		if (C->Y1>y) C->Y1=(XYZ_T)y; if (C->Y2<y) C->Y2=(XYZ_T)y;
		if (C->Z1>z) C->Z1=(XYZ_T)z; if (C->Z2<z) C->Z2=(XYZ_T)z;

		if (argc==5) cell[y][x]=(c<PV)?(Cell)c:PV;
	    }
	    if (argc==5) {
		(void)sprintf(path,"%s/%s",argv[4],paths[z]+len);
		StoreImageFile(path,Nx,Ny,bps,cell,NULL);
	    }
	}
	for (C=cluster, c=0; c<Nc; c++, C++)
	    (void)printf("%" LL_FORMAT "u\t%u\t%u\t%u\t%u\t%u\t%u\n",
			 (unsigned long long)C->pels,
			 (unsigned)C->X1,(unsigned)C->Y1,(unsigned)C->Z1,
			 (unsigned)C->X2,(unsigned)C->Y2,(unsigned)C->Z2);

	return 0;
}
