
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

#ifndef	LABEL_T
#define LABEL_T	unsigned
#endif
#ifndef	PELS_T
#define PELS_T	unsigned long
#endif
#ifndef	XYZ_T
#define XYZ_T	unsigned short
#endif

extern void	Error(),ReadSliceImage();

static LABEL_T	Trace(link,old)
LABEL_T		*link,old;
{
	LABEL_T	new=link[old];

	return (new==0)?old:(link[old]=Trace(link,new));
}

#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,path[LEN];
	int	Nx,Ny,Nz,BPS,x,y,z;
	Cell	**value,v;

	struct Last {
		Cell	v;
                LABEL_T	l;
	} **last;

	LABEL_T	Nl,*link,label,l,lz,ly,lx,Nc;

	struct Cluster {
		union {
			LABEL_T	l;
                        Cell	v;
		} u;
		PELS_T	pels;
		double	x0,y0,z0;
		XYZ_T	x1,y1,z1,x2,y2,z2;
        } *cluster,*c;

	size_t	len;

#ifdef	IOM
	Cell	***image;
#endif
	if (argc!=3 && argc!=4)
	    Error("usage : si_pcs directory nameFile {directory}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,&BPS);

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

#ifdef	IOM
	if ((image=ALLOC(Cell **,Nz))==NULL ||
	    (*image=ALLOC(Cell *,(size_t)Nz*(size_t)Ny))==NULL ||
	    (**image=ALLOC(Cell,(size_t)Nz*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for image.");

	value=image[0];
	for (y=1; y<Ny; y++) value[y]=value[y-1]+Nx;
	for (z=1; z<Nz; z++) {
	    value=image[z]=image[z-1]+Ny;
	    for (y=0; y<Ny; y++) value[y]=value[y-1]+Nx;
	}
	for (z=0; z<Nz; z++) ReadSliceImage(paths[z],Nx,Ny,image[z]);

#define READ(z)		value=image[z]
#else
	if ((value=ALLOC(Cell *,Ny))==NULL ||
	    (*value=ALLOC(Cell,(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for value.");

	for (y=1; y<Ny; y++) value[y]=value[y-1]+Nx;

#define READ(z)	ReadSliceImage(paths[z],Nx,Ny,value)
#endif
	if ((last=ALLOC(struct Last *,1+Ny))==NULL ||
	    (*last=ALLOC(struct Last,(size_t)Nx+
				     (size_t)Ny*(size_t)(1+Nx)))==NULL)
	    Error("no allocatable memory for last.");

	++last;
	for (y=0; y<Ny; y++) last[y]=last[y-1]+1+Nx;

#define VX	last[y][x-1].v
#define VY	last[y-1][x].v
#define VZ	last[y][x].v

	for (x=0; x<Nx; x++) last[-1][x].v=0;
	for (y=0; y<Ny; y++) last[y][-1].v=0;

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) last[y][x].v=0;

	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if ((v=value[y][x])>VZ && v>VY && v>VX &&
		    ++Nl>(LABEL_T)(~0)/sizeof(LABEL_T))
		    Error("too many links.");

		last[y][x].v=v; 
	    }
	}
	if (Nl==2) Error("no non-zero value area.");

	if ((link=ALLOC(LABEL_T,Nl))==NULL)
	    Error("no allocatable memory for link.");

	link[0]=link[1]=0;

#define LX	last[y][x-1].l
#define LY	last[y-1][x].l
#define LZ	last[y][x].l

#define TRACE(LABEL)	Trace(link,LABEL)
#define RESET(LABEL)	if ((label=LABEL)>1) link[label]=1

	for (x=0; x<Nx; x++) last[-1][x].l=0;
	for (y=0; y<Ny; y++) last[y][-1].l=0;

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    last[y][x].v=0; last[y][x].l=0;
	}
	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if ((v=value[y][x])==0)
		    l=0;
		else if (v<VZ || v<VY || v<VX) {
		    l=0; if (v>=VX) RESET(LX);
			 if (v>=VY) RESET(TRACE(LY));
			 if (v>=VZ) RESET(TRACE(LZ));
		}
		else if (v==VZ && v==VY && v==VX)
		    if ((lz=TRACE(LZ))>1)
			if ((ly=TRACE(LY))>1)
			    if ((lx=LX)>1) {
				l=lz; if (ly!=lz) link[ly]=lz;
				      if (lx!=lz) link[lx]=lz;
			    }
			    else
				l=link[lz]=link[ly]=1;
			else {
			    l=link[lz]=1; RESET(LX);
			}
		    else {
			l=1; RESET(LX); RESET(TRACE(LY));
		    }
		else if (v==VZ && v==VY) {
		    if ((lz=TRACE(LZ))>1)
			if ((ly=TRACE(LY))>1) {
			    l=lz; if (ly!=lz) link[ly]=lz;
			}
			else
			    l=link[lz]=1;
		    else {
			l=1; RESET(TRACE(LY));
		    }
		    RESET(LX);
		}
		else if (v==VZ && v==VX) {
		    if ((lz=TRACE(LZ))>1)
			if ((lx=LX)>1) {
			    l=lz; if (lx!=lz) link[lx]=lz;
			}
			else
			    l=link[lz]=1;
		    else {
			l=1; RESET(LX);
		    }
		    RESET(TRACE(LY));
		}
		else if (v==VY && v==VX) {
		    if ((ly=TRACE(LY))>1)
			if ((lx=LX)>1) {
			    l=ly; if (lx!=ly) link[lx]=ly;
			}
			else
			    l=link[ly]=1;
		    else {
			l=1; RESET(LX);
		    }
		    RESET(TRACE(LZ));
		}
		else if (v==VZ) {
		    l=TRACE(LZ); RESET(LX); RESET(TRACE(LY));
		}
		else if (v==VY) {
		    l=TRACE(LY); RESET(LX); RESET(TRACE(LZ));
		}
		else if (v==VX) {
		    l=LX; RESET(TRACE(LY)); RESET(TRACE(LZ));
		}
		else {
		    link[l=Nl++]=0;
		    RESET(LX); RESET(TRACE(LY)); RESET(TRACE(LZ));
		}
		last[y][x].v=v; last[y][x].l=l;
	    }
        }
	for (Nc=l=2; l<Nl; l++)
	    if (link[l]!=0)
		link[l]=TRACE(link[l]);
	    else
		if (++Nc>(LABEL_T)(~0)/sizeof(struct Cluster))
		    Error("too many clusters.");

	(void)printf("%" LL_FORMAT "u\n",(unsigned long long)(Nc-2));

	if ((cluster=ALLOC(struct Cluster,Nc-2))==NULL)
	    Error("no allocatable memory for cluster.");

	cluster-=2;

#define L	u.l
#define V	u.v

	for (Nc=l=2; l<Nl; l++) if (link[l]==0) cluster[link[l]=Nc++].L=l;

	for (c=cluster+Nc-1, l=Nl-1; l>1; l--)
	    if (c>cluster+1 && c->L==l) (c--)->V=0; else link[l]=link[link[l]];

	if (argc==4) len=strlen(argv[1])+1;

	for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    last[y][x].v=0; last[y][x].l=0;
	}
	for (Nl=2, z=0; z<Nz; z++) {
	    READ(z);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
		if ((l=((v=value[y][x])==0 || v<VZ || v<VY || v<VX)?0:
		       (v==VZ)?LZ:(v==VY)?LY:(v==VX)?LX:link[Nl++])>1)
		    if ((c=cluster+l)->V==0) {
			c->V=v;
			c->pels=1; c->x0=(double)x; c->x1=c->x2=(XYZ_T)x;
				   c->y0=(double)y; c->y1=c->y2=(XYZ_T)y;
				   c->z0=(double)z; c->z1=c->z2=(XYZ_T)z;
		    }
		    else {
			if (++(c->pels)==1) c->pels=0;

			c->x0+=(double)x; if (c->x1>x) c->x1=(XYZ_T)x; else
					  if (c->x2<x) c->x2=(XYZ_T)x;
			c->y0+=(double)y; if (c->y1>y) c->y1=(XYZ_T)y; else
					  if (c->y2<y) c->y2=(XYZ_T)y;
			c->z0+=(double)z; if (c->z1>z) c->z1=(XYZ_T)z; else
					  if (c->z2<z) c->z2=(XYZ_T)z;
		    }
		else
		    if (argc==4) value[y][x]=0;

		last[y][x].v=v; last[y][x].l=l;
	    }
	    if (argc==4) {
		(void)sprintf(path,"%s/%s",argv[3],paths[z]+len);
		StoreImageFile(path,Nx,Ny,BPS,value,NULL);
	    }
	}
	for (c=cluster+(l=2); l<Nc; l++, c++)
	    if (c->V==0)
		(void)fputs("-0\t"
			    "-0\t-0\t-0\t-0\t"
			    "-0\t-0\t-0\t-0\t-0\t-0\n",stdout);
	    else {
		(void)printf("%lu\t",(unsigned long)c->V);

		if (c->pels==0)
		    (void)fputs("-0\t-0\t-0\t-0\t",stdout);
		else
		    (void)printf("%" LL_FORMAT "u\t%lf\t%lf\t%lf\t",
				 (unsigned long long)c->pels,
				 c->x0/(double)c->pels,
				 c->y0/(double)c->pels,
				 c->z0/(double)c->pels);

		(void)printf("%u\t%u\t%u\t%u\t%u\t%u\n",
			     (unsigned)c->x1,(unsigned)c->y1,(unsigned)c->z1,
			     (unsigned)c->x2,(unsigned)c->y2,(unsigned)c->z2);
	    }

	return 0;
}
