
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t limit,char *pv);

#ifndef	UMA
#define UMA	1024	/* unit of memory allocation */
#endif

static long	*Link;

static long	Trace(long old)
{
	long	new=Link[old];

	return (new<0)?old:(Link[old]=Trace(new));
}

typedef struct {
		long	l_p;
		int	x1,y1,x2,y2;
	} Cluster;

static int	Compare(Cluster *C1,Cluster *C2)
{
	long	p1=Link[C1->l_p],p2=Link[C2->l_p];

	return (p1==p2)?0:(p1<p2)?-1:1;
}

#define PV	(Cell)(~0)
#define SOL	sizeof(long)

int main(int argc,char **argv)
{
	int	X,Y,x,y,bps;
	Cell	**cell;
	char	pv[(size_t)PV+1],
		*mae="memory allocation error.",
		*tlc="too large cluster.",
		*mrae="memory re-allocation error.";
	long	*last,L,l,m,N;
	size_t	S;
	Cluster	*C;

	if (argc!=3 && argc!=4) Error("usage : t_mcl_1 TIFF rangeList {TIFF}");

	ReadImageFile(argv[1],&X,&Y,NULL,&cell,NULL);

	RangeList(argv[2],PV,pv);

	if ((last=(long *)malloc(SOL*(X+1)))==NULL ||
	    (Link=(long *)malloc(SOL*(S=UMA)))==NULL) Error(mae);

	++last;

	Link[0]=0; L=1; for (x=(-1); x<X; x++) last[x]=0;

#define CHECK(size)     if ((size)>=0) Error(tlc)

	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) {
	    if (!pv[cell[y][x]]) {
		CHECK(--Link[l=0]);
	    }
	    else if ((l=last[x])!=0) {
		CHECK(--Link[l=Trace(l)]);
		if ((m=last[x-1])!=0 && m!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
	    }
	    else if ((l=last[x-1])!=0) {
                CHECK(--Link[l]);
	    }
	    else {
		if (L==S && (Link=(long *)
			     realloc(Link,SOL*(S+=UMA)))==NULL) Error(mrae);

		Link[l=L++]=(-1);
	    }
	    last[x]=l;
	}
	if (L<S && (Link=(long *)realloc(Link,SOL*L))==NULL) Error(mrae);

	for (N=l=1; l<L; l++) if ((m=Link[l])<0) ++N; else Link[l]=Trace(m);

	if ((C=(Cluster *)malloc(sizeof(Cluster)*N))==NULL) Error(mae);

	for (N=l=1; l<L; l++) if (Link[l]<0) C[N++].l_p=l;

	if (N>1) qsort(C+1,N-1,sizeof(Cluster),(int (*)())Compare);

	C[0].l_p=0;
	for (m=0; m<N; m++) {
	    C[m].l_p=Link[l=C[m].l_p]; Link[l]=(-m);

	    C[m].x1=X; C[m].y1=Y; C[m].x2=C[m].y2=(-1); 
	}
	for (l=1; l<L; l++) if ((m=Link[l])>0) Link[l]=(-Link[m]);
	for (l=1; l<L; l++) if ((m=Link[l])<0) Link[l]=(-m);

	L=1; for (x=0; x<X; x++) last[x]=0;

	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) {
	    if (!pv[cell[y][x]])
		l=0;
	    else if ((l=last[x])==0 && (l=last[x-1])==0)
		l=L++;

	    m=Link[l]; if (argc==4) cell[y][x]=(m<PV)?m:PV;

	    if (C[m].x1>x) C[m].x1=x; if (C[m].x2<x) C[m].x2=x;
	    if (C[m].y1>y) C[m].y1=y; if (C[m].y2<y) C[m].y2=y;

	    last[x]=l;
	}
	if (argc==4) {
	    for (bps=1; bps<sizeof(Cell)*8 && (size_t)1<<bps<N; bps<<=1) ;

	    StoreImageFile(argv[3],X,Y,bps,cell,NULL);
	}
	(void)printf("%ld\n",N);
	for (m=0; m<N; m++)
	    (void)printf("%lu\t%d\t%d\t%d\t%d\n",
			 (unsigned long)(-(long long)C[m].l_p),
			 C[m].x1,C[m].y1,C[m].x2,C[m].y2);

	return 0;
}
