
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

#ifndef	LABEL_T
#define LABEL_T	unsigned
#endif
#ifndef	PELS_T
#define PELS_T	unsigned long
#endif
#ifndef	XY_T
#define XY_T	unsigned short
#endif

extern void	Error(),RangeList();

static LABEL_T	Trace(link,old)
LABEL_T		*link,old;
{
	LABEL_T	new=link[old];

	return (new==0)?old:(link[old]=Trace(link,new));
}

typedef struct {
		PELS_T	pels;
		union {
			LABEL_T	l;
			struct {
				LABEL_T	old,new;
			} LABEL;
			struct {
				XY_T	x1,y1,x2,y2;
			} XY;
		} UNION;
	} Cluster;

static int	Compare(C1,C2)
Cluster		*C1,*C2;
{
	return (C1->pels==C2->pels)?0:(C1->pels>C2->pels)?-1:1;
}

#define PV	(Cell)(~0)
#define BPS	(sizeof(Cell)<<3)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,x,y,bps,noa=1;
	Cell	**cell;
	char	pv01[(size_t)PV+1];
	LABEL_T	*last,Nl,l,*link,lx,Nc,c;
	Cluster	*cluster,*C;

	if (argc!=3 && argc!=4) Error("usage : t_mhl TIFF rangeList {TIFF}");

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	RangeList(argv[2],(size_t)PV,pv01);

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

	if ((last=ALLOC(LABEL_T,1+Nx))==NULL)
	    Error("no allocatable memory for last.");

	++last; for (x=(-1); x<Nx; x++) last[x]=1;

	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if ((cell[y][x]=pv01[cell[y][x]])!=0)
		l=noa=0;
	    else {
		if (!last[x] && !last[x-1] &&
		    ++Nl>(LABEL_T)(~0)/sizeof(LABEL_T))
		    Error("too many links.");

		l=1;
	    }
	    last[x]=l;
	}
	if (noa) Error("no object area.");

	if ((link=ALLOC(LABEL_T,Nl))==NULL)
	    Error("no allocatable memory for link.");

	link[0]=link[1]=0;

	for (x=0; x<Nx; x++) last[x]=1;

	for (Nl=2, y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) {
		if (cell[y][x])
		    l=0;
		else if ((l=last[x])!=0) {
		    l=Trace(link,l);
		    if ((lx=last[x-1])!=0 && lx!=l)
			if (lx==1) l=link[l]=1; else link[lx]=l;
		}
		else if ((l=last[x-1])==0)
		    link[l=Nl++]=0;

		last[x]=l;
	    }
	    if (y>0 && Nx>1 && (l=last[Nx-1])>1) link[l]=1;
	}
	if (Ny>1)
	    for (x=Nx-2; x>=1; x--)
		if ((l=last[x])!=0 && (l=Trace(link,l))!=1) link[l]=1;

	for (Nc=l=2; l<Nl; l++)
	    if (link[l]!=0)
		link[l]=Trace(link,link[l]);
	    else
		if (++Nc>(LABEL_T)(~0)/sizeof(Cluster))
		    Error("too many clusters.");

	(void)printf("%" LL_FORMAT "u\n",(unsigned long long)Nc);

	if ((cluster=ALLOC(Cluster,Nc))==NULL)
	    Error("no allocatable memory for cluster.");

#define L	UNION.l

	for (Nc=l=0; l<Nl; l++)
	    if (link[l]==0) {
		C=cluster+(link[l]=Nc++); C->pels=0; C->L=l;
	    }

	for (C=cluster+Nc-1, l=Nl-1; l>1; l--)
	    if (l==C->L) --C; else link[l]=link[link[l]];

	for (x=0; x<Nx; x++) last[x]=1;

	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if (cell[y][x])
		c=0;
	    else if ((c=last[x  ])==0 &&
		     (c=last[x-1])==0) c=link[Nl++];

	    if ((last[x]=c)<=1) c=1-c;

	    if (++cluster[c].pels==0) --cluster[c].pels;
	}

#define OLD	UNION.LABEL.old
#define NEW	UNION.LABEL.new

	if (Nc>2) {
	    for (C=cluster+(c=2); c<Nc; c++, C++) C->OLD=c;

	    qsort(cluster+2,(size_t)(Nc-2),sizeof(Cluster),Compare);

	    for (C=cluster+(c=2); c<Nc; c++, C++) cluster[C->OLD].NEW=c;

	    for (l=2; l<Nl; l++)
		if (link[l]!=1) link[l]=cluster[link[l]].NEW;
	}

#define X1	UNION.XY.x1
#define Y1	UNION.XY.y1
#define X2	UNION.XY.x2
#define Y2	UNION.XY.y2

	for (C=cluster, c=0; c<Nc; c++, C++) {
	    C->X1=(XY_T)(Nx-1); C->Y1=(XY_T)(Ny-1); C->X2=C->Y2=0;
	}
	if (argc==4) for (bps=1; bps<BPS && 1<<bps<Nc; bps<<=1) ;

	for (x=0; x<Nx; x++) last[x]=1;

	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if (cell[y][x])
		c=0;
	    else if ((c=last[x  ])==0 &&
		     (c=last[x-1])==0) c=link[Nl++];

	    if ((last[x]=c)<=1) c=1-c;

	    C=cluster+c;
	    if (C->X1>x) C->X1=(XY_T)x; if (C->X2<x) C->X2=(XY_T)x;
	    if (C->Y1>y) C->Y1=(XY_T)y; if (C->Y2<y) C->Y2=(XY_T)y;

	    if (argc==4) cell[y][x]=(c<PV)?(Cell)c:PV;
	}
	if (argc==4) StoreImageFile(argv[3],Nx,Ny,bps,cell,NULL);

	for (C=cluster, c=0; c<Nc; c++, C++)
	    (void)printf("%" LL_FORMAT "u\t%u\t%u\t%u\t%u\n",
			 (unsigned long long)C->pels,
			 (unsigned)C->X1,(unsigned)C->Y1,
			 (unsigned)C->X2,(unsigned)C->Y2);

	return 0;
}
