
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

#ifndef	LABEL_T
#define LABEL_T	unsigned
#endif
#ifndef	PELS_T
#define PELS_T	unsigned long
#endif
#ifndef	XY_T
#define XY_T	unsigned short
#endif

extern void	Error();

static LABEL_T	Trace(link,old)
LABEL_T		*link,old;
{
	LABEL_T	new=link[old];

	return (new==0)?old:(link[old]=Trace(link,new));
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,BPS,x,y;
	Cell	**value,v;

	struct Last {
		Cell	v;
		LABEL_T	l;
	} *last;

	LABEL_T	Nl,*link,label,l,ly,lx,Nc;

	struct Cluster {
		union {
			LABEL_T	l;
			Cell	v;
		} u;
		PELS_T	pels;
		double	x0,y0;
		XY_T	x1,y1,x2,y2;
	} *cluster,*c;

	if (argc!=2 && argc!=3) Error("usage : t_pcs TIFF {TIFF}");

	ReadImageFile(argv[1],&Nx,&Ny,&BPS,&value,NULL);

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

	if ((last=ALLOC(struct Last,1+Nx))==NULL)
	    Error("no allocatable memory for last.");

	++last;

#define VX	last[x-1].v
#define VY	last[x].v

	for (x=(-1); x<Nx; x++) last[x].v=0;

	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if ((v=value[y][x])>VY && v>VX &&
		++Nl>(LABEL_T)(~0)/sizeof(LABEL_T)) Error("too many links.");

	    last[x].v=v; 
	}
	if (Nl==2) Error("no non-zero value area.");

	if ((link=ALLOC(LABEL_T,Nl))==NULL)
	    Error("no allocatable memory for link.");

	link[0]=link[1]=0;

#define LX	last[x-1].l
#define LY	last[x].l

#define TRACE(LABEL)	Trace(link,LABEL)
#define RESET(LABEL)	if ((label=LABEL)>1) link[label]=1

	last[-1].l=0;
	for (x=0; x<Nx; x++) {
	    last[x].v=0; last[x].l=0;
	}
	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if ((v=value[y][x])==0)
		l=0;
	    else if (v<VY || v<VX) {
		l=0; if (v>=VX) RESET(LX);
		     if (v>=VY) RESET(TRACE(LY));
	    }
	    else if (v==VY && v==VX)
		if ((ly=TRACE(LY))>1)
		    if ((lx=LX)>1) {
			l=ly; if (lx!=ly) link[lx]=ly;
		    }
		    else
			l=link[ly]=1;
		else {
		    l=1; RESET(LX);
		}
	    else if (v==VY) {
		l=TRACE(LY); RESET(LX);
	    }
	    else if (v==VX) {
		l=LX; RESET(TRACE(LY));
	    }
	    else {
		link[l=Nl++]=0; RESET(LX); RESET(TRACE(LY));
	    }
	    last[x].v=v; last[x].l=l;
	}
	for (Nc=l=2; l<Nl; l++)
	    if (link[l]!=0)
		link[l]=TRACE(link[l]);
	    else
		if (++Nc>(LABEL_T)(~0)/sizeof(struct Cluster))
		    Error("too many clusters.");

	(void)printf("%" LL_FORMAT "u\n",(unsigned long long)(Nc-2));

	if ((cluster=ALLOC(struct Cluster,Nc-2))==NULL)
	    Error("no allocatable memory for cluster.");

	cluster-=2;

#define L	u.l
#define V	u.v

	for (Nc=l=2; l<Nl; l++) if (link[l]==0) cluster[link[l]=Nc++].L=l;

	for (c=cluster+Nc-1, l=Nl-1; l>1; l--)
	    if (c>cluster+1 && c->L==l) (c--)->V=0; else link[l]=link[link[l]];

	for (x=0; x<Nx; x++) {
	    last[x].v=0; last[x].l=0;
	}
	for (Nl=2, y=0; y<Ny; y++) for (x=0; x<Nx; x++) {
	    if ((l=((v=value[y][x])==0 || v<VY
				       || v<VX)?0:(v==VY)?LY:
						  (v==VX)?LX:link[Nl++])>1)
		if ((c=cluster+l)->V==0) {
		    c->V=v;
		    c->pels=1; c->x0=(double)x; c->x1=c->x2=(XY_T)x;
			       c->y0=(double)y; c->y1=c->y2=(XY_T)y;
		}
		else {
		    if (++(c->pels)==1) c->pels=0;

		    c->x0+=(double)x; if (c->x1>x) c->x1=(XY_T)x; else
				      if (c->x2<x) c->x2=(XY_T)x;
		    c->y0+=(double)y; if (c->y1>y) c->y1=(XY_T)y; else
				      if (c->y2<y) c->y2=(XY_T)y;
		}
	    else
		if (argc==3) value[y][x]=0;

	    last[x].v=v; last[x].l=l;
	}
	if (argc==3) StoreImageFile(argv[2],Nx,Ny,BPS,value,NULL);

	for (c=cluster+(l=2); l<Nc; l++, c++)
	    if (c->V==0)
		(void)fputs("-0\t-0\t-0\t-0\t-0\t-0\t-0\t-0\n",stdout);
	    else {
		(void)printf("%lu\t",(unsigned long)c->V);

		if (c->pels==0)
		    (void)fputs("-0\t-0\t-0\t",stdout);
		else
		    (void)printf("%" LL_FORMAT "u\t%lf\t%lf\t",
				 (unsigned long long)c->pels,
				 c->x0/(double)c->pels,
				 c->y0/(double)c->pels);

		(void)printf("%u\t%u\t%u\t%u\n",
			     (unsigned)c->x1,(unsigned)c->y1,
			     (unsigned)c->x2,(unsigned)c->y2);
	    }

	return 0;
}
